/*
 * Copyright (C) 2023 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"
#include "features.hpp"
#include "calls.hpp"

#include <moderncli/keyfile.hpp>
#include <moderncli/tasks.hpp>

#include <algorithm>
#include <csignal>

using namespace server;
namespace {
task_queue tq;

void feature_echo(const feature_t& feature, const event_t& event) {
    auto& digits = feature->extra;
    auto& user = feature->origin;
    auto from = std::to_string(user.line());

    if(!digits.empty())
        system_message("Echo digits = " + digits, from);
    logger.debug(2, "echo request from {}, digits=\"{}\"", from, digits);
    event.terminate_call();
}

void feature_reload([[maybe_unused]] const feature_t& feature, const event_t& event) {
    event.terminate_call();
    raise(SIGHUP);
}

void feature_dialer(const feature_t& feature, const event_t& event) {
    auto& user = feature->origin;
    auto& invite = feature->server;
    auto up = user.shm();
    auto display = user.contact().display();

    // make sure we have something to invite...
    if(!is(invite) || !invite.allow_invite()) {
        event.fail_call(SIP_NOT_FOUND);
        return;
    }

    // Make sure we have capacity...
    if(++up->count > up->lines) {
        --up->count;
        event.fail_call(SIP_BUSY_HERE);
        return;
    }


    // check for busy...
    auto ep = invite.shm();
    if(++ep->count >  ep->lines) {
        --ep->count;
        event.fail_call(SIP_BUSY_HERE);
        return;
    }
    logger.debug(4, "call from \"{}\" for feature {}", display, feature->name);

    auto call_id = session::next_callid();
    if(!call_id) {
        --up->count;
        event.fail_call(SIP_BUSY_EVRYWHERE);
    }

    session::call_server(call_id, event, feature);
}

using feat_t = void (*)(const feature_t& feature, const event_t& event);

std::unordered_map<std::string, feat_t> features;
std::unordered_map<std::string, std::string> used;
std::unordered_map<std::string, feat_t> names = {
    {"echo", &feature_echo},
    {"reload", &feature_reload},
};

void reload(const keyfile& keys) {
    const auto codes = keys.at("features");
    features.clear();
    used.clear();
    std::for_each(codes.begin(), codes.end(), [](const auto& pair) {
        const auto& [code, name] = pair;
        if(!name.empty() && name[0] == '@') {
            features[code] = &feature_dialer;
            used[code] = name.c_str() + 1;
        }
        else {
            auto exec = names[name];
            if(exec == nullptr) return;
            features[code] = exec;
            used[code] = name;
        }
    });
    logger.debug(4, "loaded {} feature handler(s)", features.size());
}
} // end anon namespace

void feature::startup(const keyfile& config) {
    logger.info("starting features thread");

    tq.errors([](const std::exception& e) {
        logger.error("feature error: {}", e.what());
    });

    tq.shutdown([]{
        logger.info("stopping features thread");
    });

    ::reload(config);
    tq.startup();
}

void feature::reload(const keyfile& config) {
    tq.dispatch([config] {
        ::reload(config);
    });
}

void feature::dialed(std::shared_ptr<const feature> feat, const event_t &event) {
    tq.dispatch([event, feat] {
        const auto& dialed = feat->code;
        for(const auto& pair : features) {
            const auto& [dial, exec] = pair;
            auto len = dial.length();
            if(dialed.length() >= len && dialed.substr(0, len) == dial) {
                feat->name = used[dial];
                if (dialed.length() > len)
                    feat->extra = dialed.substr(len);
                exec(feat, event);
                return;
            }
        };
        event.fail_call(SIP_NOT_FOUND);
    });
}

void feature::shutdown() {
    tq.shutdown();
}

