/*
 * Copyright (C) 2021 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"
#include "messages.hpp"
#include "events.hpp"
#include "utils.hpp"

#include <moderncli/keyfile.hpp>
#include <moderncli/filesystem.hpp>
#include <moderncli/datetime.hpp>
#include <moderncli/tasks.hpp>
#include <moderncli/scan.hpp>

#include <list>

using namespace server;
namespace {
time_t max_expires = 30;
std::ofstream logfile, cdrfile;
task_queue tq;

void reload(const keyfile& keys) {
    auto config = keys.at("messages");
    max_expires = get_seconds_or(config["expires"], 300U);

    // Allows for log rotation...
    if(logfile.is_open())
        logfile.close();
    logfile.open(SERVER_LOGGER_PATH "/coventry.log", std::ios_base::app);

    if(cdrfile.is_open())
        cdrfile.close();
    cdrfile.open(SERVER_LOGGER_PATH "/calls.log", std::ios_base::app);
}
} // end anon namespace

message::message(std::string_view from_name, std::string_view to_uri, std::string_view route_to, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context) : data_(std::make_shared<data>(from_name, to_uri, route_to, msg_subject, msg_text, msg_type)), ctx_(context) {
    time(&data_->expires);
    data_->expires += max_expires;
}

message::message(int dialog, int msg_status, int msg_reason, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context) : data_(std::make_shared<data>(msg_status, msg_reason, msg_subject, msg_text, msg_type)), ctx_(context), did_(dialog) {
    time(&data_->expires);
    data_->expires += max_expires;
}

void message::startup(const keyfile& config) {
#ifdef DEBUG
    fsys::remove(SERVER_LOGGER_PATH "/coventry.log");
    fsys::remove(SERVER_LOGGER_PATH "/calls.log");
#endif

    tq.errors([](const std::exception& e) {
        logger.error("message error: {}", e.what());
    });

    tq.shutdown([]{
        logger.info("stopping message thread");
    });

    ::reload(config);
    logger.info("starting messages thread");
    tq.startup();
}

void message::reload(const keyfile& config) {
    tq.dispatch([config] {
        ::reload(config);
    });
}

void message::post_log(const logmsg& log) {
    tq.dispatch([log] {
        if(!logfile.is_open()) return;
        const auto now = local_time(log.created);
        logfile << std::put_time(&now, ISO_DATETIME) << " " << log.msg << std::endl;
    });
}

void message::post_cdr(const logmsg& cdr) {
    tq.dispatch([cdr] {
        if(!cdrfile.is_open()) return;
        const auto duration = format("{}d{}h{}m{}s", cdr.duration / 86400, (cdr.duration / 3600) % 24, (cdr.duration / 60) % 60, cdr.duration % 60);
        std::stringstream json;
        json <<  std::put_time(gmtime(&cdr.created), "{\"m\":\"cdr\",\"v\":0,\"p\":\"sip\",\"s\":\"coventry/" PROJECT_VERSION "\",\"zt\":\"%FT%TZ\",") << cdr.msg << R"(,"td":")" << duration  << "\"}";
        cdrfile << json.str() << std::endl;
    });
}

void message::notify_local(const message& send) {
    tq.dispatch([send] {
        const auto& context = send.context();
        const auto& did = send.did();
        const auto& type = send.type();
        const auto& body = send.body();
        const auto& event = send.subject();
        const auto& reason = send.reason();
        auto status = send.status();

        logger.debug(4, "deliver to {}", send.route());

        std::this_thread::yield();

        osip_message_t *msg = nullptr;
        const exosip::guard_t lock(context);
        auto result = eXosip_insubscription_build_notify(context, did, status, reason, &msg);
        if(msg) {
            if(!event.empty())
                osip_message_set_header(msg, "Event", event.c_str());
            osip_message_set_body(msg, body.data(), body.size());
            osip_message_set_content_type(msg, type.c_str());
            eXosip_insubscription_send_request(context, did, msg);
        }
        else
            logger.error("failed to send message to {}; error={}", send.to(), result);
    });
}

void message::deliver_local(const message& send) {
    tq.dispatch([send] {
        const auto& context = send.context();
        const auto& to = send.to();
        const auto& type = send.type();
        const auto& body = send.body();
        const auto& from = send.from();
        const auto& subject = send.subject();

        logger.debug(4, "deliver to {}", send.route());

        std::this_thread::yield();

        osip_message_t *msg = nullptr;
        const exosip::guard_t lock(context);
        auto result = eXosip_message_build_request(context, &msg, "MESSAGE", to.c_str(), from.c_str(), send.route());
        if(msg) {
            if(!subject.empty())
                osip_message_set_subject(msg, subject.c_str());
            osip_message_set_body(msg, body.data(), body.size());
            osip_message_set_content_type(msg, type.c_str());
            allow_local(msg);
            eXosip_message_send_request(context, msg);
        }
        else
            logger.error("failed to send message to {}; error={}", send.to(), result);
    });
}

void message::shutdown() {
    tq.shutdown();
}

void server::logger_notify(const std::string& msg, const char *type) {
    if(eq(type, "debug")) return;
    message::post_log(logmsg(msg));
}

