/*
 * Copyright (C) 2021 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MESSAGES_HPP_
#define MESSAGES_HPP_

#include "server.hpp"        // IWYU pragma: keep
#include "exosip.hpp"

#include <moderncli/strings.hpp>

namespace server {
struct logmsg final {
    logmsg() = delete;

    explicit logmsg(const std::string& text, time_t started = 0) :
    msg(std::move(text)), duration(-1) {
        if(started) {
            time(&duration);
            created = started;
            duration -= created;
        }
        else
            time(&created);
    };

    std::string msg;
    time_t created;
    time_t duration;
};

class message final {
public:
    message(std::string_view from_name, std::string_view to_uri, std::string_view route_to, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context);

    message(int dialog, int status, int reason, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context);

    message() = default;

    auto from() const {
        return ctx_ ? data_->from : "";
    }

    auto to() const {
        return ctx_ ? data_->to : "";
    }

    auto subject() const {
        return ctx_ ? data_->subject : "";
    }

    auto route() const {
        return ctx_ ? data_->route.c_str() : "";
    }

    auto text() const {
        return ctx_ ? data_->text.c_str() : "";
    }

    auto size() const {
        return str_size(text(), 160);
    }

    auto context() const {
        return ctx_;
    }

    auto type() const {
        return ctx_ ? data_-> type : "text/plain";
    }

    auto body() const {
        return std::string_view(text());
    }

    auto did() const {
        return did_;
    }

    auto status() const {
        return did_ != -1 ? data_->status : EXOSIP_SUBCRSTATE_UNKNOWN;
    }

    auto reason() const {
        return did_ != -1 ? data_->reason : DEACTIVATED;
    }

    static void startup(const keyfile& config);
    static void reload(const keyfile& config);
    static void shutdown();
    static void deliver_local(const message& send);
    static void notify_local(const message& send);
    static void post_log(const logmsg& log);
    static void post_cdr(const logmsg& cdr);

private:
    struct data {
        data(const std::string_view& from_name, const std::string_view& to_uri, const std::string_view& route_uri, const std::string_view& msg_subject, const std::string_view& msg_text, const std::string_view& msg_type) :
        from{from_name}, to{to_uri}, route{route_uri}, subject{msg_subject}, text{msg_text}, type{msg_type} {}

        data(int msg_status, int msg_reason, const std::string_view& msg_subject, const std::string_view& msg_text, const std::string_view& msg_type) :
        subject{msg_subject}, text{msg_text}, type{msg_type}, status{msg_status}, reason{msg_reason} {}

        std::string from, to, route, subject, text, type;
        time_t expires{};
        int status{}, reason{};
    };

    std::shared_ptr<data> data_;
    eXosip_t *ctx_{nullptr};
    int did_{-1};
};

void logger_notify(const std::string& msg, const char *type);
} // end namespace
#endif
