/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef	REGISTRY_HPP_
#define REGISTRY_HPP_

#include "server.hpp"       // IWYU pragma: keep
#include "nonce.hpp"
#include "ipc.hpp"
#include "exosip.hpp"

namespace server {
using keys_t = std::unordered_map<std::string, std::string>;
class event_t;
class registry;

inline constexpr auto registry_entries = 80U;
inline constexpr auto line_min = 10U;
inline constexpr auto line_max = line_min + registry_entries - 1U;
inline constexpr auto extension_min = 100U;
inline constexpr auto extension_max = 999U;
inline constexpr auto expires_year = 60L * 60L * 24L * 365L;

class contact_t final {
public:
    contact_t() = default;
    explicit contact_t(const osip_uri_t *uri);

    explicit contact_t(const event_t& event) {
        from(event);
    }

    explicit operator bool() const {
        return !host_.empty();
    }

    auto operator!() const {
        return host_.empty();
    }

    auto operator==(const contact_t& other) const {
        return host_ == other.host_ && port_ == other.port_ && user_ == other.user_;
    }

    auto operator!=(const contact_t& other) const {
        return host_ != other.host_ || port_ != other.port_ || user_ != other.user_;
    }

    auto port() const {return port_;}
    auto user() const {return user_;}
    auto host() const {return host_;}
    auto display() const {return display_;}

    auto empty() const {
        return host_.empty();
    }

    auto id() const {
        auto result(host_);
        if(host_.empty())
            return result;
        if(port_)
            result += ":" + std::to_string(port_);
        if(user_.empty())
            return result;
        return user_ + "@" + result;
    }

    auto addr() const {
        auto result(host_);
        if(host_.empty())
            return result;
        if(port_)
            result += ":" + std::to_string(port_);
        return result;
    }

    void from(const event_t& event);

protected:
    friend class registry;

    std::string host_;
    uint16_t port_{};
    std::string user_;
    std::string display_;
};

class registry final {
public:
    registry() : ptr_(nullptr) {}
    explicit registry(unsigned key);
    registry(const registry&) = default;

    explicit operator eXosip_t *() const {
        return context_;
    }

    explicit operator bool() const;
    auto operator!() const -> bool;

    auto operator==(const registry& from) {
        return get() == from.get();
    }

    auto presence() const {
        if(!shm_ && operator bool())
            return pbx_reg::HERE;
        if(!shm_)
            return pbx_reg::GONE;
        return shm_->presence;
    }

    auto open() const {
        return operator bool() && (presence() == pbx_reg::HERE || presence() == pbx_reg::AWAY);
    }

    auto line() const {
        return ptr_ ? ptr_->line : 0;
    }

    auto value() const {
        return ptr_ ? ptr_->line : 0;
    }

    auto expires() const {
        return ptr_ ? ptr_->expires : 0;
    }

    auto schema() const {
        return ptr_ ? ptr_->schema.c_str() : "";
    }

    auto context() const {
        return context_;
    }

    auto hash() const {
        return std::size_t(value());
    }

    auto use_count() const {
        return ptr_.use_count();
    }

    auto empty() const {
        return !ptr_;
    }

    auto nonce() const {
        return ptr_ ? ptr_->nonce : nonce_t();
    }

    auto allow_invite() const {
        return shm_ ? shm_->flags.invite && open() : false;
    }

    auto allow_coverage() const {
        return shm_ ? shm_->flags.invite && operator bool() && presence() == pbx_reg::HERE : false;
    }

    auto allow_message() const {
        return shm_ ? shm_->flags.message && open() : false;
    }

    auto welcome() const {
        auto result = false;
        if(ptr_ && shm_) {
            result = shm_->flags.message && ptr_->welcome;
            ptr_->welcome = false;
        }
        return result;
    }

    auto user_id() const {
        return ptr_ ? ptr_->contact.user() : "";
    }

    auto display() const {
        return ptr_ ? ptr_->contact.display() : "";
    }

    auto remote() const {
        return ptr_ ? origin().addr() : "";
    }

    auto caller() const {
        return ptr_ && context_ ? ptr_-> caller : "";
    }

    auto agent() const {
        return ptr_ && context_ ? ptr_-> agent : "unknown";
    }

    auto type() const {
        return ptr_ && context_ ? ptr_-> type : "unknown";
    }

    auto acl() const {
        return ptr_ && context_ ? ptr_-> acl : std::vector<std::string>();
    }

    auto location() const {
        return ptr_ && context_ ? ptr_-> location : "unknown";
    }

    auto shm() const {
        return shm_;
    }

    auto local_uri() const -> std::string;
    auto local_alias(std::string_view to) const -> std::string;
    auto local_to() const -> std::string;
    auto origin_uri() const -> std::string;
    auto route_to() const -> std::string;
    auto from(const event_t& event) const -> std::string;
    auto to(const event_t& event) const -> std::string;
    auto origin() const -> contact_t;
    auto contact() const -> contact_t;
    auto keys() const -> server::keys_t;
    void keys(const server::keys_t& pairs);
    auto refresh(const event_t& event, time_t period) -> bool;
    void receipt(const event_t& event, time_t period) const;
    auto verify(const event_t& event) -> bool;
    void drop();
    void allow(exosip::msg_t msg);

    static auto is_line(std::string_view text) -> bool;
    static auto is_extension(std::string_view text) -> bool;
    static auto is_group(std::string_view text) -> bool;
    static auto is_active() -> bool;

    static void startup(const keyfile& config, unsigned level);
    static void reload(const keyfile& config);
    static void shutdown();
    static void pbx_message(const pbx_msg& msg);
    static void pbx_action(const pbx_msg& ipc);
    static void feature_code(const event_t& event);
    static void release_user(const event_t& event);
    static void refresh_user(const event_t& event);
    static void option_user(const event_t& event);
    static void publish_user(const event_t& event);
    static void message_user(const event_t& event);
    static void invite_user(const event_t& event);

private:
    struct data_t final {
        std::mutex mutex;
        server::keys_t keys;

        volatile time_t expires;
        unsigned line;
        contact_t contact, origin;
        std::string schema;
        nonce_t nonce;
        std::string caller;
        std::string client;
        std::string agent;
        std::string type;
        std::string location;
        std::vector<std::string> acl;
        uint32_t count;
        bool welcome;
    };

    std::shared_ptr<data_t> ptr_;
    eXosip_t *context_{nullptr};
    pbx_reg *shm_{nullptr};

    auto get() const -> data_t * {
        return ptr_.get();
    }
};

void change_level(unsigned level);
void system_message(std::string_view text, std::string_view to = "all", std::string_view subject = "");
} // end namespace

namespace std { // NOLINT
template<>
struct hash<server::registry> {
    auto operator()(const server::registry& key) const {
        return key.hash();
    }
};
} // end namespace
#endif
