/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"
#include "events.hpp"
#include "features.hpp"
#include "messages.hpp"
#include "registry.hpp"
#include "calls.hpp"
#include "ipc.hpp"

#include <moderncli/args.hpp>
#include <moderncli/keyfile.hpp>
#include <moderncli/socket.hpp>
#include <moderncli/filesystem.hpp>

#include <csignal>

#include <sys/stat.h>

#ifdef  SYSTEMD_SERVICE
#include <systemd/sd-daemon.h>
#else
inline auto sd_notify(int, const char *) {  // NOLINT
    return 0;
}
#endif

using namespace server;

namespace { // private data and functions
const auto AGENT_ID = SERVER_ID "/" PROJECT_VERSION;
fsys::path config_file = SERVER_CONFIG_FILE;
fsys::path prefix_path = SERVER_PREFIX_PATH;
std::string status = "READY=1\nSTATUS=unknown"; // set on reload
Socket ipc;

args::option opt_ipv6('6',  "--ipv6", "enable ipv6");
args::option opt_config('c',"--config", "configuration file", "file");
args::option helpflag('h',  "--help", "display this list");
args::option opt_prefix(0,  "--prefix", "prefix directory", "path");
args::option verbose('v',   "--verbose", "verbosity, repeat for debug", true);
args::option version(0,     "--version", "display version information");
args::option althelp('?',   nullptr, nullptr);

std::string shutdown_message = "Shutting down...";
exosip::context_t local(AGENT_ID);
exosip::context_t remote(AGENT_ID);

auto reload() {
    auto config = keyfile::create()
        .load("server", {
            {"sitename", "Coventry Server"},
            {"location", "unspecified"},
            {"weather", "unknown"},
            {"webserver", "80"},
            {"urlschema", "http"},
            {"hostname", system_hostname()},
            {"address", "*"},
            {"remote", "*"},
        })
        .load("messages", {
            {"welcome", "Welcome to Coventry"},
            {"shutdown", "Shutting down..."},
        })
        .load("calls", {
            {"mode", "proxy"},
            {"ring", "4"},
            {"delayed", "12"},
        })
        .load("common", {
            {"lines", "1"},
            {"presence", "here"},
            {"type", "generic"},
            {"location", "unspecified"},
            {"cabling", "unspecified"},
        });

    // initialize [groups] and [access] so no error occurs
    config["extensions"];
    config["display"];
    config["groups"];
    config["access"];
    config["rooms"];
    config["zones"];
    config["trunk"];

    // load composite config
    config.load(config_file);
    config.load("dynamic.conf");
    config.load("custom.conf");
    config.load("state.conf");

    // process state symlink
    std::string state = "default";
    if(fsys::exists("state.conf") && fsys::is_symlink("state.conf")) {
        auto symlink = fsys::read_symlink("state.conf");
        if(symlink.filename() == symlink)
            state = symlink.stem().string();
    }

    // set run state
    config["server"]["state"] = state;
    logger.debug(3, "current state {}", state);
    status = "READY=1\nSTATUS=" + state;

    // validate config
    const auto& hostname = config["server"]["hostname"];
    auto& realm = config["server"]["realm"];
    auto& algorithm = config["server"]["algorithm"];
    auto& names = config["server"]["names"];
    auto& upstream = config["remote"]["upstream"];

    auto family = lower_case(config["server"]["family"]);
    if(family == "af_inet6" || family == "inet6" || family == "ipv6" || family == "6")
        ipv6 = true;

    if(!upstream.empty()) {
        if(upstream.substr(0, 4) != "sip:")
            logger.crit(98, "{}: invalid upstream uri", upstream);
    }

    if(algorithm.empty())
        algorithm = "SHA-256";

    algorithm = upper_case(algorithm);
    if(algorithm != "MD5" && algorithm != "SHA-256") {
        print(stderr, "*** warning: {}: invalid\n", algorithm);
        algorithm = "SHA-256";
    }

    if(realm.empty())
        realm = hostname;

    // alternate names we might answer for...
    if(names.empty()) {
        if(hostname != "localhost" && hostname != "localhost.localdomain") {
            auto pos = hostname.find_first_of('.');
            names = hostname.substr(0, pos) + ".local";
            if(pos != std::string::npos)
                names += "," + hostname.substr(0, pos);
            names += ",localhost,localhost.localdomain";
        }
        else if(hostname == "localhost.localdomain")
            names = "localhost";
    }
    names = hostname + "," + names;

    if(!config.exists("features")) {
        // Default features if no feature plan...
        config["features"];
        config["features"]["*99"] = "echo";
        config["features"]["*98"] = "reload";
        config["features"]["*97"] = "@weather";
    }

    running = true;
    return config;
}

void msgs(const pbx_msg& msg) {
    if(msg.ver != PBX_VERSION) {
        logger.error("ipc version invalid");
        return;
    }
    switch(msg.type) {
    case PBX_LEVEL:
        change_level(msg.body.level);
        break;
    case PBX_MESSAGE:
        registry::pbx_message(msg);
        break;
    case PBX_RELOAD:
        registry::pbx_message(msg);
        kill(getpid(), SIGHUP);
        break;
    case PBX_ACTIVE:
    case PBX_ASSIGN:
    case PBX_CLEARS:
    case PBX_PRESENCE:
        registry::pbx_action(msg);
        break;
    default:
        logger.debug(4, "received unknown message {}", static_cast<int>(msg.type));
    }
}

void control() {
    pbx_msg msg{};
    fsys::remove(SERVER_CONTROL_IPC);
    ipc.bind(address_t(SERVER_CONTROL_IPC), SOCK_DGRAM);
    if(ipc)
        logger.info("ipc control attached {}", SERVER_CONTROL_IPC);
    else {
        logger.error("ipc control failed; err={}", ipc.err());
        return;
    }

    for(;;) {
        if(!recv(ipc, msg)) {
            logger.info("ipc control exiting; reason={}", ipc.err());
            return;
        }
        msgs(msg);
    }
}

void stop() {
    ipc.release();
    sd_notify(1, "STOPPING=1");
    running = false;
    system_message(shutdown_message);
    registry::shutdown();
    fsys::remove(SERVER_CONTROL_IPC);
}

void cleanup() {
    running = false;
    fsys::remove(SERVER_CONTROL_IPC);
}
} // anon namespace

auto main(int argc, const char **argv) -> int {
    auto exit_code = 0;
    on_crit(cleanup);

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({SERVER_ID " [options]"}, {
                "coventry " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Start Coventry Service"
            });
            return 0;
        }

        if(is(opt_config))
            config_file = *opt_config;

        if(is(opt_prefix))
            prefix_path = fsys::canonical(*opt_prefix);

        if(is(version)) {
            print("Version: {}\nConfig: {}\nPrefix: {}\n", PROJECT_VERSION,
                config_file, std::string{prefix_path.u8string()});
            return 0;
        }

        if(is(opt_ipv6))
            ipv6 = true;

        umask(007);
        fsys::current_path(prefix_path);

        // create ipc info
        fsys::remove("ipc.json");
        std::ofstream ipc("ipc.json");
        ipc << R"({"ipc_path":")" << SERVER_IPCPATH << R"(",)" << std::endl
            << R"( "control_path":")" << SERVER_CONTROL_IPC << R"(",)" << std::endl
            << R"( "msg_size":)" << sizeof(pbx_msg) << "," << std::endl
            << R"( "sys_size":)" << sizeof(pbx_sys) << "," << std::endl
            << R"( "reg_size":)" << sizeof(pbx_reg) << "," << std::endl
            << R"( "call_size":)" << sizeof(pbx_call) << "," << std::endl
            << R"( "reg_count":)" << registry_entries << "," << std::endl
            << R"( "call_count":)" << TUNE_CALL_LIMIT << std::endl
            << "}" << std::endl;
        ipc.close();

        // setup logging
        logger.set(verbose.count(), logger_notify);
        system_logger::open(SERVER_ID, LOG_INFO);

        std::set_new_handler([]() {
            crit(9, "*** coventry: out of memory\n");
        });

        // protect signals from other threads
        sigset_t sigs;
        sigemptyset(&sigs);
        sigaddset(&sigs, SIGALRM);
        sigaddset(&sigs, SIGINT);
        sigaddset(&sigs, SIGHUP);
        sigaddset(&sigs, SIGKILL);
        sigaddset(&sigs, SIGTERM);
        pthread_sigmask(SIG_BLOCK, &sigs, nullptr);

        // start services
        auto config = reload();
        shutdown_message = config.at("messages").at("shutdown");

        // start threads
        session::startup(config);
        feature::startup(config);
        message::startup(config);
        registry::startup(config, verbose.count());
        std::thread local_thread(exec_local, std::ref(local), config);
        std::thread remote_thread(exec_remote, std::ref(remote), config);
        std::thread control_thread(control);
        std::this_thread::sleep_for(std::chrono::milliseconds(20));

        auto timer = sync_clock() + std::chrono::seconds(60);
        alarm(60);
        logger.info("started server");
        sd_notify(0, status.c_str());
        while(running) {
            int sig{};
            if(sigwait(&sigs, &sig))
                crit(-1, "Broken sigwait");
            switch(sig) {
            case SIGINT:
                std::cout << std::endl;  // for c/r after ^C
                stop();
                break;
            case SIGALRM:
                // sync alarm every 60 seconds, used for stats...
                sd_notify(0, "WATCHDOG=1");
                logger.debug(3, "minute timer wakeup");
                timer += std::chrono::seconds(60);
                alarm(std::chrono::duration_cast<std::chrono::seconds>(timer - sync_clock()).count());
                break;
            case SIGTERM:
            case SIGQUIT:
                stop();
                break;
            case SIGHUP:
                sd_notify(0, "RELOADING=1");
                logger.info("reloading server");
                try {
                    config = reload();
                    registry::reload(config);
                    message::reload(config);
                    feature::reload(config);
                    session::reload(config);
                }
                catch(const std::exception& e) {
                    logger.error("reload: ", e.what());
                }
                catch(...) {
                    logger.error("reload: unknown error");
                }
                sd_notify(0, status.c_str());
                break;
            default:
                break;
            }
        }

        exiting = true;
        local.close();
        if(remote.active())
            remote.close();

        control_thread.join();
        local_thread.join();
        remote_thread.join();
    }
    catch(const bad_arg& e) {
        crit(-1, "Usage: {} [options]\nerror: {}\n", SERVER_ID, e.what());
    }
    catch(const std::exception& e) {
        if(running)
            logger.crit(-1, "{}", e.what());
        else
            crit(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(running)
            logger.crit(-1, "unknown failure");
        else
            crit(-1, "error: unknown failure\n");
    }

    logger.info("stopped server; reason={}", exit_code);
    cleanup();
    return exit_code;
}
