/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "utils.hpp"
#include "events.hpp"

#include <moderncli/strings.hpp>
#include <moderncli/digest.hpp>

using namespace server;
using namespace crypto;

namespace {
auto sha256_ha2(const event_t& event) {
    std::string entity;
    if(event.qop() == "auth-int") {
        digest_t sha(EVP_sha256());
        osip_body_t *body = nullptr;
        if(event.request())
            osip_message_get_body(event.request(), 0, &body);
        if(body && body->body && body->length)
            sha.update(body->body, body->length);
        sha.finish();
        entity = to_hex(sha.data(), sha.size());
    }

    digest_t sha(EVP_sha256());
    sha.update(event.method(), str_size(event.method()));
    sha.update(":", 1);
    sha.update(event.uri().c_str(), event.uri().size());
    if(event.qop() == "auth-int") {
        sha.update(":", 1);
        sha.update(entity.data(), entity.size());
    }
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto md5_ha2(const event_t& event) {
    digest_t md5(EVP_md5());
    md5.update(event.method(), str_size(event.method()));
    md5.update(":", 1);
    md5.update(event.uri().c_str(), event.uri().size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}
} // end namespace anon

auto util::sha256_secret(std::string_view userid, std::string_view realm, std::string_view secret) -> std::string {
    digest_t sha(EVP_sha256());
    sha.update(userid.data(), userid.size());
    sha.update(":", 1);
    sha.update(realm.data(), realm.size());
    sha.update(":", 1);
    sha.update(secret.data(), secret.size());
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto util::sha256_digest(std::string_view secret, const event_t& event) -> std::string {
    auto ha2 = sha256_ha2(event);
    auto nonce = event.nonce();

    digest_t sha(EVP_sha256());
    sha.update(secret.data(), secret.size());
    sha.update(":", 1);
    sha.update(nonce.data(), nonce.size());
    sha.update(":", 1);

    if(event.qop() == "auth" || event.qop() == "auth-int") {
        sha.update(event.count().data(), event.count().size());
        sha.update(":", 1);
        sha.update(event.client().data(), event.client().size());
        sha.update(":", 1);
        sha.update(event.qop().data(), event.qop().size());
        sha.update(":", 1);
    }

    sha.update(ha2.data(), ha2.size());
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto util::md5_secret(std::string_view userid, std::string_view realm, std::string_view secret) -> std::string {
    digest_t md5(EVP_md5());
    md5.update(userid.data(), userid.size());
    md5.update(":", 1);
    md5.update(realm.data(), realm.size());
    md5.update(":", 1);
    md5.update(secret.data(), secret.size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}

auto util::md5_digest(std::string_view secret, const event_t& event) -> std::string {
    auto ha2 = md5_ha2(event);
    auto nonce = event.nonce();

    digest_t md5(EVP_md5());
    md5.update(secret.data(), secret.size());
    md5.update(":", 1);
    md5.update(nonce.data(), nonce.size());
    md5.update(":", 1);
    md5.update(ha2.data(), ha2.size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}

