/*
 * Copyright (C) 2022 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"       // IWYU pragma: keep
#include "ipc.hpp"

#include <moderncli/strings.hpp>
#include <moderncli/process.hpp>
#include <moderncli/socket.hpp>
#include <moderncli/array.hpp>
#include <moderncli/args.hpp>
#include <moderncli/scan.hpp>

#include <sys/stat.h>

namespace {
args::option helpflag('h',  "--help",       "display this list");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option althelp('?',   nullptr, nullptr);

class ipc_error final : public std::exception {
public:
    explicit ipc_error(const char *text, int code) noexcept :
    msg_(text), err_(code) {}

    auto what() const noexcept -> const char * override {
        return msg_;
    }

    auto exit() const noexcept {
        return err_;
    }

private:
    const char *msg_;
    int err_;
};

auto set_string(char *cp, std::size_t max, const char *from) -> std::size_t {
    assert(cp != nullptr);
    assert(from != nullptr);
    assert(max > 0);

    auto count = 0UL;
    while(--max && *from) {
        *(cp++) = *(from++);
        ++count;
    }
    *cp = 0;
    return count;
}
} // end namespace

auto main(int argc, const char **argv) -> int {
    using namespace process;

    try {
        const auto& list = args::list();
        auto count = args::parse(argc, argv);
        auto line = 0;

        if(is(helpflag) || is(althelp) || count < 1 || count > 2) {
            args::help({"pbx-presence line [presence]"}, {
                "pbx-presence " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Examine or set pbx presence for a line"
            });
            return 0;
        }

        line = get_integer(list[0], 10, 89);
        if(count == 2) {
            auto presence = lower_case(std::string{list[1]});
            pbx_msg msg{.type = PBX_PRESENCE, .ver = PBX_VERSION};
            msg.body.registry.line = line;
            set_string(msg.body.registry.value, sizeof(msg.body.registry.value) - 1, presence.c_str());

            auto ipc(Socket(AF_UNIX, SOCK_DGRAM));
            if(!ipc) throw ipc_error("control failed", 4);
            if(!send(ipc, msg, address_t(SERVER_CONTROL_IPC))) throw ipc_error("send failed", 4);
        }
        else {  // count == 1
            process::map_t rmem(SERVER_IPCPATH ".registry");
            if(!is(rmem)) throw ipc_error("registry mapping failed", 4);
            auto rmax = (rmem.size() - sizeof(pbx_sys)) / sizeof(pbx_reg);
            const auto reg = span<pbx_reg>(*rmem, rmax);
            const auto sys = void_ptr<const pbx_sys>(*rmem, reg.size_bytes());

            if(line < sys->first || line >= static_cast<int>(sys->first + rmax)) throw ipc_error("offline", 1);
            const auto& user = reg[line - sys->first];
            auto presence = "here";
            switch(user.presence) {
            case pbx_reg::HERE:
                if(user.count >= user.lines)
                    presence = "busy";
                else if(user.count)
                    presence = "call";
                else
                    presence = "here";
                break;
            case pbx_reg::AWAY:
                presence = "away";
                break;
            case pbx_reg::DND:
                presence = "dnd";
                break;
            default:
                break;
            }
            print("{}\n", presence);
        }
    }
    catch(const bad_arg& e) {
        if(is(quiet)) return -1;
        die(-1, "Usage: pbx-presence line [presence]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        if(is(quiet)) return e.exit();
        die(e.exit(), "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet)) return -1;
        die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet)) return -1;
        die(-1, "error: unknown error\n");
    }
    return 0;
}
