/*
 * Copyright (C) 2022 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "server.hpp"       // IWYU pragma: keep
#include "ipc.hpp"

#include <moderncli/process.hpp>
#include <moderncli/keyfile.hpp>
#include <moderncli/strings.hpp>
#include <moderncli/filesystem.hpp>
#include <moderncli/socket.hpp>
#include <moderncli/array.hpp>
#include <moderncli/args.hpp>
#include <moderncli/scan.hpp>

#include <sys/stat.h>

namespace {
args::option clearing('c',  "--clear",      "clear users registry keys");
args::option exists('e', "--exists",        "test line exists");
args::option invite('i', "--invite",        "line supports invite");
args::option helpflag('h',  "--help",       "display this list");
args::option message('m', "--message",      "line supports message");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option althelp('?',   nullptr, nullptr);

std::string keyname;
auto count = 0U;
auto line = 0;

class ipc_error final : public std::exception {
public:
    explicit ipc_error(const char *text, int code) noexcept :
    msg_(text), err_(code) {}

    auto what() const noexcept -> const char * override {
        return msg_;
    }

    auto exit() const noexcept {
        return err_;
    }

private:
    const char *msg_;
    int err_;
};

auto set_string(char *cp, std::size_t max, std::string_view view) -> std::size_t {
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = *(dp++);

    *cp = 0;
    return size;
}

auto set_lower(char *cp, std::size_t max, std::string_view view) -> std::size_t {
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = static_cast<char>(tolower(*(dp++)));

    *cp = 0;
    return size;
}
} // end namespace

auto main(int argc, const char **argv) -> int {
    auto exit_code = 0;
    try {
        count = args::parse(argc, argv);
        const auto& list = args::list();
        if(is(helpflag) || is(althelp) || count < 1 || count > 3 || ((is(exists) || is(message) || is(invite)) && count != 1)) {
            args::help({"pbx-registry [options] line [key [value]]"}, {
                "pbx-registry " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Examine registry, see and modify user keys"
            });
            return 0;
        }

        fsys::current_path(SERVER_PREFIX_PATH);
        keyfile keys({SERVER_CONFIG_FILE, "dynamic.conf", "state.conf", "custom.conf", "registry.keys"});
        line = get_integer(list[0], 10, 89);
        if(!keys.exists(std::to_string(line))) throw error(format("{}: undefined line", line));

        // test flags...
        if(is(invite) || is(message)) {
            process::map_t rmem(SERVER_IPCPATH ".registry");
            if(!is(rmem)) throw error("registry mapping failed");
            auto rmax = (rmem.size() - sizeof(pbx_sys)) / sizeof(pbx_reg);
            const auto reg = span<pbx_reg>(*rmem, rmax);
            const auto sys = void_ptr<const pbx_sys>(*rmem, reg.size_bytes());
            if(line < sys->first || line >= static_cast<int>(sys->first + rmax)) throw ipc_error("offline", 1);

            const auto& user = reg[line - sys->first];
            if(!user.flags.message && is(message)) throw ipc_error("unsupported", 2);
            if(!user.flags.invite && is(invite)) throw ipc_error("unsupported", 2);
            return 0;
        }

        if(is(exists)) return 0;
        pbx_msg msg{.type = PBX_ACTIVE, .ver = PBX_VERSION};
        msg.body.registry.line = line;
        if(count > 1)
            keyname = lower_case(std::string{list[1]});
        if(is(clearing))
            msg.type = PBX_CLEARS;
        else if(count == 3) {
            set_lower(msg.body.registry.key, sizeof(msg.body.registry.key) - 1, list[1]);
            set_string(msg.body.registry.value, sizeof(msg.body.registry.value) - 1, list[2]);
            msg.type = PBX_ASSIGN;
        }

        auto ipc(Socket(AF_UNIX, SOCK_DGRAM));
        if(!ipc) throw ipc_error("control failed", 4);
        if(!send(ipc, msg, address_t(SERVER_CONTROL_IPC))) throw ipc_error("send failed", 4);
        if(!is(quiet)) {
            if(is(clearing))
                print("user {} keys cleared\n", line);
            else if(count == 3)
                print("user {} key set\n", line);
            else if(count == 1)
                print("user {} valid\n", line);
        }
        if(count == 2) {
            auto user = keys.at(std::to_string(line));
            const auto& common = keys["common"];
            auto value = std::string();
            try {
                value = user.at(keyname);
            }
            catch(...) {
                try {
                    value = common.at(keyname);
                }
                catch(...) {}   // NOLINT
            }
            print("{}\n", value);
        }
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-registry [options] line [key [value]]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        if(is(quiet)) return e.exit();
        die(e.exit(), "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet)) return -1;
        die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet)) return -1;
        die(-1, "error: unknown error\n");
    }
    return exit_code;
}
